@@ -4,13 +4,14 @@ import logging
 import os
 from datetime import datetime, timedelta
 from itertools import islice
-from time import sleep
+from time import sleep, time
 from typing import Dict, List, Tuple
 
 import requests
 
 from yojenkins.monitor import BuildMonitor
 from yojenkins.utility import utility
+from yojenkins.utility.utility import fail_out, print2
 from yojenkins.yo_jenkins.jenkins_item_classes import JenkinsItemClasses
 from yojenkins.yo_jenkins.status import BuildStatus
 
@@ -19,13 +20,14 @@ logger = logging.getLogger()
 
 
 class Build():
-    """TODO Build"""
+    """Buld class"""
 
     def __init__(self, rest, auth) -> None:
         """Object constructor method, called at object creation
 
         Args:
-            None
+            rest: Rest object
+            auth: Auth object
 
         Returns:
             None
@@ -55,27 +57,21 @@ class Build():
             request_url = f"{build_url.strip('/')}/api/json"
             build_info = self.rest.request(request_url, 'get', is_endpoint=False)[0]
             if not build_info:
-                logger.debug('Failed to get build info')
-                return {}
+                fail_out('Failed to get build info')
         else:
             if not job_name and not job_url:
-                logger.debug('Failed to pass parameters that describe the build')
-                return {}
+                fail_out('No job name, job url, and build url provided')
 
             if job_name and not job_url:
                 job_url = utility.name_to_url(self.rest.get_server_url(), job_name)
 
             job_info, _, success = self.rest.request(f'{job_url.strip("/")}/api/json', 'get', is_endpoint=False)
             if not success:
-                logger.debug(f'Failed to find job info: {job_url}')
-                return {}
+                fail_out(f'Failed getting build info, because failed to request job info: {job_url}')
 
             # Check if found item type/class is a build
             if job_info['_class'] not in JenkinsItemClasses.JOB.value['class_type']:
-                logger.debug(
-                    f'Failed. The passed job information does not match a job type/class. Failed to match type/class. This item is "{job_info["_class"]}"'
-                )
-                return {}
+                fail_out(f'Failed to match job type/class. The found item is "{job_info["_class"]}"')
 
             job_last_build_number = job_info['lastBuild']['number'] if 'lastBuild' in job_info else 0
 
@@ -87,26 +83,22 @@ class Build():
                 )
                 build_number = job_last_build_number
             elif not build_number and not latest:
-                logger.debug('Failed to specify build. No build number passed and --latest flag not set')
-                return {}
+                fail_out('Failed to specify build. No build number passed and --latest flag not set')
             else:
                 # Build number is passed
                 if build_number > job_last_build_number:
-                    logger.debug('Failed to specify build. Build number exceeds last build number for this job')
-                    return {}
+                    fail_out('Failed to specify build. Build number exceeds last build number for this job')
 
             logger.debug(f'Getting build info for job "{job_info["fullName"]}, build {build_number} ...')
             build_info, _, success = self.rest.request(f'{job_url.strip("/")}/{build_number}/api/json',
                                                        'get',
                                                        is_endpoint=False)
             if not success:
-                logger.debug('Failed to request build information')
-                return {}
+                fail_out('Failed to request build info')
 
         # Check if found item type/class is a build
         if build_info['_class'] not in JenkinsItemClasses.BUILD.value['class_type']:
-            logger.debug(f'Failed to match type/class. This item is "{build_info["_class"]}"')
-            return {}
+            fail_out(f'Build found, but failed to match build type/class. This item is "{build_info["_class"]}"')
 
         # Add additional derived information
         if 'timestamp' in build_info:
@@ -128,8 +120,8 @@ class Build():
                     build_info['resultText'] = BuildStatus.RUNNING.value
                     build_info['durationFormatted'] = None
                     build_info['endDatetime'] = None
-                    build_info['elapsedFormatted'] = str(
-                        timedelta(seconds=((datetime.utcnow().timestamp()) - build_info['timestamp'] / 1000)))[:-3]
+                    build_info['elapsedFormatted'] = str(timedelta(seconds=(time() -
+                                                                            build_info['timestamp'] / 1000)))[:-3]
 
             else:
                 build_info['resultText'] = BuildStatus.UNKNOWN.value
@@ -184,34 +176,33 @@ class Build():
                                latest=latest)
 
         # If nothing is returned, check if job is queued on server
-        if not build_info:
-            logger.debug('The specified build was not found in job')
-            logger.debug('Looking for build in the server build queue ...')
-            if build_url:
-                job_url = utility.build_url_to_other_url(build_url)
-            elif job_name:
-                pass
-            elif job_url:
-                pass
-            else:
-                logger.debug('Failed to find build status text. Specify build url, job name, or job url')
-                return ""
-            logger.debug(f'Job name: {job_name}')
+        logger.debug('The specified build was not found in job')
+        logger.debug('Looking for build in the server build queue ...')
+        if build_url:
+            job_url = utility.build_url_to_other_url(build_url)
+        elif job_name:
+            pass
+        elif job_url:
+            pass
+        else:
+            fail_out('Failed to find build status text. Specify build url, job name, or job url')
+        logger.debug(f'Job name: {job_name}')
 
-            # Requesting all queue and searching queue (NOTE: Could use Server object)
-            queue_all = self.rest.request('queue/api/json', 'get')[0]
-            logger.debug(f"Number of queued items: {len(queue_all['items'])}")
-            queue_matches = utility.queue_find(queue_all, job_name=job_name, job_url=job_url)
-            if not queue_matches:
-                return {}, 0
-            queue_info = queue_matches[0]
+        # Requesting all queue and searching queue (NOTE: Could use Server object)
+        logger.debug(f'Requesting all build queue items ...')
+        queue_all = self.rest.request('queue/api/json', 'get')[0]
+        logger.debug(f"Number of queued items found: {len(queue_all['items'])}")
+        queue_matches = utility.queue_find(queue_all, job_name=job_name, job_url=job_url)
+        if not queue_matches:
+            fail_out('Failed to find running or queued builds')
+        queue_info = queue_matches[0]
 
-            if not queue_info:
-                logger.debug('Build for job NOT found in queue')
-                return BuildStatus.NOT_FOUND.value
-            else:
-                logger.debug(f'Build for job found in queue. Queue number {queue_info["id"]}')
-                return BuildStatus.QUEUED.value
+        if not queue_info:
+            logger.debug('Build for job NOT found in queue')
+            return BuildStatus.NOT_FOUND.value
+        else:
+            logger.debug(f'Build for job found in queue. Queue number {queue_info["id"]}')
+            return BuildStatus.QUEUED.value
 
         # FIXME: resultText is returned in build info. Maybe move queue check to build_info??
 
@@ -242,20 +233,16 @@ class Build():
             logger.debug(f'Build URL passed: {build_url}')
             url = build_url
         else:
-            logger.debug('NO build URL passed. Getting build information ...')
+            logger.debug('No build URL passed. Getting build information ...')
             # Get build info request
             build_info = self.info(build_url, job_name, job_url, build_number, latest)
-            if not build_info:
-                logger.debug('Failed to abort build. Build does not exist or may be queued')
-                return 0
             url = build_info['url']
 
         # Making a direct request using the passed url
         logger.debug(f'Aborting build: {url} ...')
         request_url = f"{url.strip('/')}/stop"
         if not self.rest.request(request_url, 'post', is_endpoint=False)[2]:
-            logger.debug('Failed to abort build. Build may not exist or is queued')
-            return 0
+            fail_out('Failed to abort build. Build may not exist or is queued')
 
         logger.debug('Successfully aborted build')
 
@@ -280,20 +267,16 @@ class Build():
             logger.debug(f'Build URL passed: {build_url}')
             url = build_url
         else:
-            logger.debug('NO build URL passed. Getting build information ...')
+            logger.debug('No build URL passed. Getting build information ...')
             # Get build info request
             build_info = self.info(build_url, job_name, job_url, build_number, latest)
-            if not build_info:
-                logger.debug('Build does not exist or may be queued')
-                return 0
             url = build_info['url']
 
         # Making a direct request using the passed url
         logger.debug(f'Deleting build: {url} ...')
         request_url = f"{url.strip('/')}/doDelete"
         if not self.rest.request(request_url, 'post', is_endpoint=False)[2]:
-            logger.debug('Failed to delete build. Build may not exist or is queued')
-            return 0
+            fail_out('Failed to delete build. Build may not exist or is queued')
 
         return utility.build_url_to_build_number(build_url=url)
 
@@ -316,12 +299,9 @@ class Build():
 
         # TODO: Pass a list of build numbers
         if not build_url:
-            logger.debug('NO build URL passed. Getting build information ...')
+            logger.debug('No build URL passed. Getting build information ...')
             # Get build info request
             build_info = self.info(build_url, job_name, job_url, build_number, latest)
-            if not build_info:
-                logger.debug('Build does not exist or may be queued')
-                return [], []
             build_url = build_info['url']
 
         # Making a direct request using the passed url
@@ -329,16 +309,14 @@ class Build():
         request_url = f"{build_url.strip('/')}/wfapi/describe"
         return_content, _, return_success = self.rest.request(request_url, 'get', is_endpoint=False)
         if not return_success or not return_content:
-            logger.debug('Failed to get build stages. Build may not exist, is queued, or is not a staged build')
-            return [], []
+            fail_out('Failed to get build stages. This may not be a staged build')
 
         # Getting the stage items
         # FIXME: When --url <job> and no build number is passed, it will just get the job describe, not build info
         if 'stages' in return_content:
             build_stage_list = return_content['stages']
         else:
-            logger.debug('No "stages" key found in return content. May not be a staged build')
-            return [], []
+            fail_out('Failed to find "stages" key in build info. This may not be a staged build')
 
         # Add additional derived information for each step
         for stage_info in build_stage_list:
@@ -349,7 +327,7 @@ class Build():
             stage_info['url'] = stage_info['_links']['self']['href']
 
         # Getting only the names of the stages
-        build_stage_name_list = [s['name'] for s in build_stage_list]
+        build_stage_name_list = [stage['name'] for stage in build_stage_list]
 
         return build_stage_list, build_stage_name_list
 
@@ -368,7 +346,7 @@ class Build():
         """
         # Test on build with artifacts
         return self.info(build_url=build_url, job_name=job_name, job_url=job_url,
-                         build_number=build_number)['artifacts']
+                         build_number=build_number).get('artifacts')
 
     def artifact_download(self):
         """TODO Docstring
@@ -403,12 +381,9 @@ class Build():
             logger.debug(f'Build URL passed: {build_url}')
             url = build_url
         else:
-            logger.debug('NO build URL passed. Getting build information through job ...')
+            logger.debug('No build URL passed. Getting build information through job ...')
             # Get build info request
             build_info = self.info(job_name=job_name, job_url=job_url, build_number=build_number, latest=latest)
-            if not build_info:
-                logger.debug('Build does not exist or may be queued')
-                return False
             url = build_info['url']
 
         # FIXME: Check if this is an actual build
@@ -429,8 +404,7 @@ class Build():
                                 open_file.write(chunk)
                 logger.debug('Successfully download build logs to file')
             except Exception as error:
-                logger.debug('Failed to download or save logs for build. Exception: {error}')
-                return False
+                fail_out(f'Failed to download or save logs for build. Exception: {error}')
         else:
             # Stream the logs to console
             if not follow:
@@ -440,8 +414,7 @@ class Build():
                                                                       is_endpoint=False,
                                                                       json_content=False)
                 if not return_success or not return_content:
-                    logger.debug('Failed to get console logs. Build may not exist or is queued')
-                    return False
+                    fail_out('Failed to get console logs. Build may not exist or is queued')
 
                 # If tail/last part of the log was specified
                 if tail:
@@ -454,7 +427,7 @@ class Build():
                     logger.debug(f'Only printing out last logs, lines {start_log_number} to {len(logs_list)} ...')
 
                 logger.debug('Printing out console text logs ...')
-                print(return_content)
+                print2(return_content)
             else:
                 logger.debug('Following/streaming logs from server ...')
                 logger.debug(
@@ -468,9 +441,8 @@ class Build():
                         # FIXME: Preserve already used session!  Use Rest object
                         response = requests.head(request_url, auth=auth)
                         if 'content-length' not in response.headers:
-                            logger.debug(
+                            fail_out(
                                 f'Failed to find "content-length" key in server response headers: {response.headers}')
-                            return False
                         content_length_sample_1 = int(response.headers['Content-Length'])
                         sleep(1)
                         response = requests.head(request_url, auth=auth)
@@ -494,9 +466,9 @@ class Build():
                             old_dict = new_dict
                             fetch_number += 1
 
-                            print(diff_text)
+                            print2(diff_text)
                         else:
-                            logger.debug('NO content length difference')
+                            logger.debug('No content length difference')
                 except KeyboardInterrupt:
                     logger.debug('Keyboard Interrupt (CTRL-C) by user. Stopping log following ...')
         return True
@@ -520,20 +492,17 @@ class Build():
             logger.debug(f'Build URL passed: {build_url}')
             build_url = build_url.strip('/')
         else:
-            logger.debug('NO build URL passed. Getting build information ...')
+            logger.debug('No build URL passed. Getting build information ...')
             build_info = self.info(build_url, job_name, job_url, build_number, latest)
-            if not build_info:
-                logger.debug('Build does not exist or may be queued')
-                return 0
             build_url = build_info['url']
 
         # Open the build in browser
         logger.debug(f'Opening build in web browser: "{build_url}" ...')
         success = utility.browser_open(url=build_url)
-        if success:
-            logger.debug('Successfully opened in web browser')
-        else:
-            logger.debug('Failed to open in web browser')
+        if not success:
+            fail_out('Failed to open build in web browser')
+        logger.debug('Successfully opened build in web browser')
+
         return success
 
     def monitor(self,
@@ -553,20 +522,57 @@ class Build():
         """
         if build_url:
             logger.debug(f'Build URL passed: {build_url}')
+            if not self.rest.request(f"{build_url.strip('/')}/api/json", 'head', is_endpoint=False)[0]:
+                fail_out(f'Failed to find build. The build may not exist: {build_url}')
             url = build_url
         else:
-            logger.debug('NO build URL passed. Getting build information ...')
-            # Get build info request
+            logger.debug('No build URL passed. Getting build information ...')
             build_info = self.info(build_url, job_name, job_url, build_number, latest)
-            if not build_info:
-                logger.debug('Build does not exist or may be queued')
-                return 0
             url = build_info['url']
 
         logger.debug(f'Starting monitor for: "{url}" ...')
         success = self.build_monitor.monitor_start(build_url=url, sound=sound)
-        if success:
-            logger.debug('Successfully opened monitor')
-        else:
-            logger.debug('Failed to open monitor for build')
+        if not success:
+            fail_out('Failed to start build monitor')
+        logger.debug('Successfully started build monitor')
+
         return success
+
+    def parameters(self,
+                   build_url: str = '',
+                   job_name: str = '',
+                   job_url: str = '',
+                   build_number: int = None,
+                   latest: bool = False) -> Tuple[list, list]:
+        """TODO Docstring
+
+        Args:
+            TODO
+
+        Returns:
+            TODO
+        """
+        # TODO: Pass a list of build numbers
+        if not build_url:
+            logger.debug('No build URL passed. Getting build information ...')
+            # Get build info request
+            build_info = self.info(build_url, job_name, job_url, build_number, latest)
+            build_url = build_info['url']
+
+        logger.debug(f'Getting build parameters for: "{build_url}" ...')
+
+        # Check if item has any parameter actions
+        parameter_actions = utility.get_item_action(build_info, 'hudson.model.ParametersAction')
+        if not parameter_actions:
+            fail_out('This build does not have any build parameters')
+
+        # Get the parameter definitions
+        parameters = parameter_actions[0]['parameters']
+
+        # List of parameter items
+        parameters_list = [
+            f'{parameter["_class"].split(".")[-1]} - {parameter["name"]} - {parameter["value"]}'
+            for parameter in parameters
+        ]
+
+        return parameters, parameters_list
